local Sackville = LibStub("AceAddon-3.0"):GetAddon("Sackville", "AceConsole-3.0")
local B = LibStub("LibBabble-Inventory-3.0")
local BL = B:GetLookupTable()
-- local PT = LibStub("LibPeriodicTable-3.1")

local sprintf = string.format

Sackville.Filter = {}
local allFilters = {}
local itemQualities = {
    [ITEM_QUALITY0_DESC] = 0,
    [ITEM_QUALITY1_DESC] = 1,
    [ITEM_QUALITY2_DESC] = 2,
    [ITEM_QUALITY3_DESC] = 3,
    [ITEM_QUALITY4_DESC] = 4,
    [ITEM_QUALITY5_DESC] = 5,
    [ITEM_QUALITY6_DESC] = 6,
    [ITEM_QUALITY7_DESC] = 7
}

local function AlwaysFalse()
    return false
end

local function IsArmor(item)
    return allFilters[BL["Armor"]].filter(item)
end

local function IsShirt(item)
    return item.equipLoc == "INVTYPE_BODY"
end

local function IsTrinket(item)
    return item.equipLoc == "INVTYPE_TRINKET"
end

local function IsNecklace(item)
    return (item.equipLoc == "INVTYPE_NECK")
end

local function IsRing(item)
    return (item.equipLoc == "INVTYPE_FINGER")
end

local function IsTabard(item)
    -- Sackville.Print("Comparing equipLoc " .. item.equipLoc .. " to INVTYPE_TABARD")
    return item.equipLoc == "INVTYPE_TABARD"
end

local function IsEquipment(item)
    if item.equipLoc == "" then
        return false
    else
        return true
    end
end

local function MakeQualityFilter(quality)
    local code = "return function (item) " .. string.format(      
        [[return item.quality == "%d" end]], quality)
    return assert(loadstring(code)())
end

local function MakeStandardFilter(category, subCategory)
    assert(category ~= nil)

    local code = "return function (item) "    
    if subCategory then
        code = code .. string.format([[
            return (item.type == "%s" and item.subType == "%s")]],
            category, subCategory)
    else
        code = code .. string.format([[
            return (item.type == "%s")]], category)
    end
    
    code = code .. " end"
    return assert(loadstring(code)())
end

function Sackville.Filter:AddFilter(category, catFilter, subCategory, subCatFilter)
-- category: top level category
-- catFilter: function returning true/false if the tested item is a member
-- subCategory: second level category
-- subCatFilter: function returning true/false if the tested item is a member
    if not allFilters[category] then
        allFilters[category] = {}
        allFilters[category].subCategories = {}
        allFilters[category].filter = catFilter
    end

    if subCategory then
        -- Sackville:Print("AddFilter cat=" .. category .. ", subCat=" .. subCategory)
        allFilters[category].subCategories[subCategory] = {}
        allFilters[category].subCategories[subCategory].filter = subCatFilter
    end
end

function deepcopy(t) -- not used
    if type(t) ~= 'table' then return t end
    local mt = getmetatable(t)
    local res = {}
    for k,v in pairs(t) do
        if type(v) == 'table' then
        v = deepcopy(v)
        end
        res[k] = v
    end
    setmetatable(res,mt)
    return res
end

function Sackville.Filter:ItemMatchesSection(item, section)
    if section.isDefault then
        -- Sackville:Print("* Hit default section: " .. section.name)
        return true
    end
    
    for _, catName in ipairs(section.categories) do
        local matchRules = Sackville.categories[catName]
        assert(matchRules, "nil matchRules. catName:" .. catName)
        if matchRules and #matchRules > 0 then
            local isMatch = Sackville.Filter:ItemMatchesRules(item, matchRules)
            if isMatch then return true end
        else
            Sackville:Print("* Found no rules in section:" .. section.name)        
        end
    end
    
    return false
end

function Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex)
    ruleIndex = ruleIndex or 1
    
    local rule = matchRules[ruleIndex]
    local returnValue = false
    
    if rule then
        assert(rule.operation == "and" or rule.operation == "or")
        -- Sackville:Print("- Processing rule index " .. ruleIndex)
        if matchRules[ruleIndex+1] then
        
            if rule.operation == "and" then
                if rule.subCategory then
                    local objCat = allFilters[rule.category]
                    local objSubCat = objCat.subCategories[rule.subCategory]
                    
                    returnValue = objSubCat.filter(item) and
                        Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex+1)
                    -- Sackville:Print(sprintf("- Trying filter %s/%s: %s",
                        -- rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE")) 
                else
                    local objCat = allFilters[rule.category]
                    returnValue = objCat.filter(item) and
                        Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex+1)
                    -- Sackville:Print(sprintf("- Trying filter %s/%s: %s",
                        -- rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE"))
                end
            else -- "or"
                if rule.subCategory then
                    local objCat = allFilters[rule.category]
                    local objSubCat = objCat.subCategories[rule.subCategory]
                    returnValue = objSubCat.filter(item) or
                        Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex+1)
                    -- Sackville:Print(sprintf("- Trying filter %s/%s: %s",
                        -- rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE"))
                else
                    local objCat = allFilters[rule.category]
                    returnValue = objCat.filter(item) or
                        Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex+1)
                    -- Sackville:Print(sprintf("- Trying filter %s/%s: %s",
                        -- rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE"))
                end            
            end
        else -- This is the last (and maybe only) rule
            if rule.subCategory then
                local objCat = allFilters[rule.category]
                local objSubCat = objCat.subCategories[rule.subCategory]
                returnValue = objSubCat.filter(item)
                -- Sackville:Print(sprintf("- Trying filter %s/%s: %s",
                        -- rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE"))
            else
                local objCat = allFilters[rule.category]
                returnValue = objCat.filter(item)
                -- Sackville:Print(sprintf("- Trying filter %s/%s: %s",
                        -- rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE"))
            end       
        end
        
        if rule.negate then
            Sackville:Print("+ NEGATING RETURN VALUE")
            returnValue = not returnValue
        end
    end
    

    return returnValue
end

-- These are "standard" because they match up with the Type and SubType
-- values that are returned by the game from a call to GetItemInfo()
local standardCategories = {
    [BL["Armor"]] = {
        BL["Miscellaneous"],
        BL["Cloth"],
        BL["Leather"],
        BL["Mail"],
        BL["Plate"],
        BL["Shields"],
        BL["Librams"],
        BL["Idols"],
        BL["Totems"],
        BL["Sigils"]
    },
    [BL["Consumable"]] = {
        BL["Food & Drink"],
        BL["Potion"],
        BL["Elixir"],
        BL["Flask"],
        BL["Bandage"],
        BL["Item Enhancement"],
        BL["Scroll"],
        BL["Other"]
    },
    [BL["Container"]] = {
        BL["Bag"],
        BL["Enchanting Bag"],
        BL["Engineering Bag"],
        BL["Gem Bag"],
        BL["Herb Bag"],
        BL["Mining Bag"],
        BL["Soul Bag"],
        BL["Leatherworking Bag"]
    },
    [BL["Gem"]] = {
        BL["Blue"],
        BL["Green"],
        BL["Orange"],
        BL["Meta"],
        BL["Prismatic"],
        BL["Purple"],
        BL["Red"],
        BL["Simple"],
        BL["Yellow"]
    },
    [BL["Miscellaneous"]] = {
        BL["Junk"],
        BL["Reagent"],
        BL["Pet"],
        BL["Holiday"],
        BL["Mount"],
        BL["Other"]
    },
    [BL["Recipe"]] = {
        BL["Alchemy"],
        BL["Blacksmithing"],
        BL["Book"],
        BL["Cooking"],
        BL["Enchanting"],
        BL["Engineering"],
        BL["First Aid"],
        BL["Leatherworking"],
        BL["Tailoring"]
    },
    [BL["Quest"]] = {
        BL["Quest"]
    },
    [BL["Trade Goods"]] = {
        BL["Armor Enchantment"],
        BL["Cloth"],
        BL["Devices"],
        BL["Elemental"],
        BL["Enchanting"],
        BL["Explosives"],
        BL["Herb"],
        BL["Jewelcrafting"],
        BL["Leather"],
        BL["Materials"],
        BL["Meat"],
        BL["Metal & Stone"],
        BL["Other"],
        BL["Parts"],
        BL["Trade Goods"],
        BL["Weapon Enchantment"]
    },
    [BL["Weapon"]] = {
        BL["Bows"],
        BL["Crossbows"],
        BL["Daggers"],
        BL["Guns"],
        BL["Fishing Poles"],
        BL["Fist Weapons"],
        BL["Miscellaneous"],
        BL["One-Handed Axes"],
        BL["One-Handed Maces"],
        BL["One-Handed Swords"],
        BL["Polearms"],
        BL["Staves"],
        BL["Thrown"],
        BL["Two-Handed Axes"],
        BL["Two-Handed Maces"],
        BL["Two-Handed Swords"],
        BL["Wands"]
    }
}

-- Populate allFilters with filters based on the primary and secondary
-- categories being equal to item.type and item.subType
for category, subCategories in pairs(standardCategories) do
    local catFilter = MakeStandardFilter(category)
    
    for _, subCategory in ipairs(subCategories) do
        local subCatFilter = MakeStandardFilter(category, subCategory)
        Sackville.Filter:AddFilter(category, catFilter, subCategory, subCatFilter)
    end
end

for name, value in pairs(itemQualities) do
    -- FIXME: Localize "Quality"
    Sackville.Filter:AddFilter("Quality", AlwaysFalse, name, MakeQualityFilter(value))
end

Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Armor"], IsArmor)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Trinket"], IsTrinket)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Shirt"], IsShirt)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Neck"], IsNecklace)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Ring"], IsRing)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Tabard"], IsTabard)