local Sackville = LibStub("AceAddon-3.0"):GetAddon("Sackville", "AceConsole-3.0")
Sackville.VirtualBag = {}
Sackville.VirtualBag.__index = Sackville.VirtualBag

local HEADER_HEIGHT = 15
local FOOTER_HEIGHT= 15
local BAG_PADDING_LEFT = 15
local BAG_PADDING_RIGHT = 15
local BAG_PADDING_TOP = 40
local BAG_PADDING_BOTTOM = 15
local SECTION_PADDING_LEFT = 0
local SECTION_PADDING_RIGHT = 0
local SECTION_PADDING_TOP = 15
local SECTION_PADDING_BOTTOM = 0
local MAX_WIDTH = 400
local BUTTON_SIZE = 40
local SECTION_MIN_WIDTH = BUTTON_SIZE
local SECTION_MIN_HEIGHT = BUTTON_SIZE + 20
local BAG_MIN_WIDTH = SECTION_MIN_WIDTH + BAG_PADDING_LEFT + BAG_PADDING_RIGHT
local BAG_MIN_HEIGHT = SECTION_MIN_HEIGHT + BAG_PADDING_TOP + BAG_PADDING_BOTTOM
local SECTION_PADDING = 10

local backdrop = {
  -- path to the background texture
  bgFile = "Interface/DialogFrame/UI-DialogBox-Background-Dark",  
  -- path to the border texture
  edgeFile = "Interface/DialogFrame/UI-DialogBox-Border",
  -- true to repeat the background texture to fill the frame, false to scale it
  tile = true,
  -- size (width or height) of the square repeating background tiles (in pixels)
  tileSize = 32,
  -- thickness of edge segments and square size of edge corners (in pixels)
  edgeSize = 22,
  -- distance from the edges of the frame to those of the background texture (in pixels)
  insets = {
    left = 6,
    right = 6,
    top = 6,
    bottom = 6
  }
}

local backdrop_bank = {
  -- path to the background texture
  bgFile = "Interface/DialogFrame/UI-DialogBox-Gold-Background",  
  -- path to the border texture
  edgeFile = "Interface/DialogFrame/UI-DialogBox-Border",
  -- true to repeat the background texture to fill the frame, false to scale it
  tile = true,
  -- size (width or height) of the square repeating background tiles (in pixels)
  tileSize = 32,
  -- thickness of edge segments and square size of edge corners (in pixels)
  edgeSize = 22,
  -- distance from the edges of the frame to those of the background texture (in pixels)
  insets = {
    left = 6,
    right = 6,
    top = 6,
    bottom = 6
  }
}

local backdrop_title = {
   bgFile = "Interface/WorldStateFrame/WorldStateFinalScoreFrame-TopBackground",  
   edgeFile = "Interface/DialogFrame/UI-DialogBox-Border",
   -- true to repeat the background texture to fill the frame, false to scale it
   tile = false,
   -- size (width or height) of the square repeating background tiles (in pixels)
   tileSize = 16,
   -- thickness of edge segments and square size of edge corners (in pixels)
   edgeSize = 22,
   -- distance from the edges of the frame to those of the background texture (in pixels)
   insets = {
      left = 6,
      right = 6,
      top = 6,
      bottom = 0
   }
}

local function TableHasValues(t)
    if next(t) == nil then
        return false
    else
        return true
    end
end

local function MakeMultiSorter(sortCriteria)
-- Returns a comparison function that can be used with table.sort()
-- sortCriteria is a table with key-value pairs where key is a valid name of an
-- element in an item object and value is either '<' or '>', which controls
-- whether or not we should sort by ascending or descending values.
	local code = "return function(a,b) if false then\n"

    for index=1, #sortCriteria do
        local key = sortCriteria[index].key
        local comparison = sortCriteria[index].comparison
        assert(key and comparison)
		code = code .. string.format("elseif a.%s ~= b.%s then return a.%s %s b.%s\n", key, key, key, comparison, key)
	end
    code = code .. " end end"
	
    -- Sackville:Print("MakeMultiSorter source = " .. code)
	return assert(loadstring(code)())
end

local function kpairs(t, sortCriteria)
    local a = {}
    for _, item in pairs(t) do table.insert(a, item) end
    table.sort(a, MakeMultiSorter(sortCriteria))
    local i = 0      -- iterator variable
    local iter = function ()   -- iterator function
        i = i + 1
        if a[i] == nil then return nil
        else return a[i].bagslot, a[i]
        end
    end
    return iter
end

local function spairs(t, sortCriteria)
    local a = {}
    for _, section in pairs(t) do
        table.insert(a, section) 
        -- Sackville:Print("+ spairs section:" .. section.name)
    end        
    table.sort(a, MakeMultiSorter(sortCriteria))
    local i = 0      -- iterator variable
    local iter = function ()   -- iterator function
        i = i + 1
        if a[i] == nil then return nil
        else return a[i].name, a[i]
        end
    end
    return iter
end

local function UpdateSectionDimensions(section)
    local f = section.frame
    if section.row == 0 then
        f:SetWidth(section.col * BUTTON_SIZE)
    else
        f:SetWidth(section.max_width - (section.max_width % BUTTON_SIZE))
    end
    
    f:SetHeight(section.padding_top + (section.row+1) * BUTTON_SIZE )
    f.titleFrame:SetWidth(f:GetWidth())
    -- Sackville:Print(string.format(
        -- "Updating section title width to min(%d,%d)",
        -- f:GetWidth()+(SECTION_PADDING/2), f.titleFrame.text:GetStringWidth()))
    f.titleFrame.text:SetWidth(f:GetWidth()+(SECTION_PADDING/2))

    -- Sackville:Print(string.format("UpdateSectionDimensions(%s) -> (%d,%d)",
        -- section.name, f:GetWidth(), f:GetHeight()))
    
end

local function PositionItemButton(item)
    local x_pos, y_pos
    local section = item.section

    x_pos = BUTTON_SIZE*(section.col)
       
    if (x_pos + BUTTON_SIZE  > section.max_width) then
        section.row = section.row + 1
        section.col = 0
        x_pos = 0
    end

    -- y_pos = -section.padding_top - BUTTON_SIZE*(section.row)   
    y_pos =  -section.padding_top - BUTTON_SIZE*(section.row)   
    -- Sackville:Print(string.format("Positioning %s at (%s, %s)", item.name, x_pos, y_pos))
    item.button.wrapper:SetPoint("TOPLEFT", x_pos, y_pos)
    item.button.wrapper:Show()    
    item.button:Show()
    
    section.col = section.col + 1
    
    UpdateSectionDimensions(section)
end

function Sackville.VirtualBag:New(bagOptions, displayOrder)
    -- Sackville:Print("Adding new virtual bag: " .. bagOptions.name)
    local self = {}
    setmetatable(self, Sackville.VirtualBag)
    self.name = bagOptions.name
    self.displayOrder = displayOrder
    self.type = bagOptions.bagType
    self.defaultPriority = bagOptions.priority or 0
    self.defaultSortCriteria = bagOptions.sortCriteria or {}
    self.max_width = MAX_WIDTH
    self.sections = {}
    self.padding_left = BAG_PADDING_LEFT
    self.padding_right = BAG_PADDING_RIGHT
    self.padding_top = BAG_PADDING_TOP
    self.padding_bottom = BAG_PADDING_BOTTOM
    self.next_x = self.padding_left
    self.next_y = self.padding_top
    
    self.frame = self:CreateBagFrame(self.name, true)
    self.min_width = self.frame:GetWidth()

    return self
end

function Sackville.VirtualBag:AddSection(sectionOptions, displayOrder)
    local section = {}
    section.vbag = self
    section.name = sectionOptions.name
    section.displayOrder = displayOrder
    section.priority = sectionOptions.priority or section.vbag.defaultPriority
    section.sortCriteria = sectionOptions.sortCriteria or section.vbag.defaultSortCriteria
    section.isDefault = sectionOptions.defaultSection
    section.categories = sectionOptions.categories or {}
    section.items = {}
    section.row = 0
    section.col = 0
    section.padding_top = SECTION_PADDING_TOP
    section.max_width = self.max_width - self.padding_left - self.padding_right

    section.frame = self:CreateSectionFrame(section.name)
    section.frame:Show()
    
    self.sections[section.name] = section
end

function Sackville.VirtualBag:CreateSectionFrame(sectionName)
    -- Sackville:Print(string.format("+  %s: Section(%s)", self.name, sectionName))
    local parentFrame = self:GetFrame()
    local frameName = parentFrame:GetName() .. "_" .. sectionName
    local sectionFrame = (_G[frameName]) and _G[frameName] 
        or CreateFrame('Frame', frameName, parentFrame)
    sectionFrame:Hide()
    -- sectionFrame.texture = sectionFrame:CreateTexture(frameName .. "_texture", 'BACKGROUND')
    -- sectionFrame.texture:SetTexture(0,1,0,0.75)
    -- sectionFrame.texture:SetAllPoints(sectionFrame)
    sectionFrame:SetMinResize(SECTION_MIN_WIDTH, SECTION_MIN_HEIGHT)

    -- TODO: add title frame    
    local titleFrame =    _G[frameName .. "_TitleFrame"] or 
        CreateFrame("Frame", frameName .. "_TitleFrame", sectionFrame)
    titleFrame:SetFrameStrata("HIGH")
    titleFrame:ClearAllPoints()
    titleFrame:SetPoint("TOPLEFT", 0, 0)
    titleFrame:SetHeight(13)
    titleFrame:SetWidth(sectionFrame:GetWidth())

    -- titleFrame.texture = titleFrame:CreateTexture(frameName .. "_TitleFrame_texture", 'BACKGROUND')
    -- titleFrame.texture:SetTexture(1,0,0,0.75)
    -- titleFrame.texture:SetAllPoints(titleFrame)

    titleFrame:Show()

    local titleFontString = _G[frameName .. "_TitleText"] or 
        titleFrame:CreateFontString(frameName .. "_TitleText")
    titleFontString:SetFontObject("GameFontWhite")
    titleFontString:SetText(sectionName)
    titleFontString:ClearAllPoints()
    titleFontString:SetHeight(12)
    titleFontString:SetPoint("LEFT", 0, 0)
    titleFontString:SetJustifyH("LEFT")

    sectionFrame.titleFrame = titleFrame
    titleFrame.text = titleFontString    
    
    return sectionFrame
end

function Sackville.VirtualBag:CreateBagFrame(bagName, moveable)
    -- Sackville:Print("CreateBagFrame: " .. bagName)
    local frameName = "SackvilleVB_" .. bagName
    local f = (_G[frameName]) and _G[frameName] or CreateFrame('Frame', frameName, UIParent)

    f:Hide()

    f:SetBackdrop(backdrop)
    f:SetHeight(BAG_MIN_HEIGHT)

    f:SetFrameStrata('HIGH')
    
    if (moveable) then
        f:SetMovable(true)
        f:EnableMouse(true)
        f:RegisterForDrag('LeftButton')
        f:SetScript('OnMouseDown', function(f, button)
         if button ~= 'LeftButton' then return end
         f.pOrigin = f.pOrigin or {}
         f.pOrigin.point, f.pOrigin.anchor, f.pOrigin.relPoint, f.pOrigin.x, f.pOrigin.y = f:GetPoint(1)
         f:StartMoving()
         f.pStart = f.pStart or {}
         f.pStart.x, f.pStart.y = select(4, f:GetPoint(1))
        end)
        f:SetScript('OnMouseUp', function(f, button)
         if button ~= 'LeftButton' then return end
         f.pEnd = f.pEnd or {}
         f.pEnd.x, f.pEnd.y = select(4, f:GetPoint(1))
         f:StopMovingOrSizing()
         f:ClearAllPoints()
         local x = f.pOrigin.x + (f.pEnd.x - f.pStart.x)
         local y = f.pOrigin.y + (f.pEnd.y - f.pStart.y)
         f:SetPoint(f.pOrigin.point, f.pOrigin.anchor, f.pOrigin.relPoint, x, y)
        end)
    end


    local titleFrame =    _G[frameName .. "_TitleFrame"] or 
        CreateFrame("Frame", frameName .. "_TitleFrame", f)
    titleFrame:ClearAllPoints()
    titleFrame:SetPoint("TOPLEFT", 0, 0)
    titleFrame:SetHeight(35)
    titleFrame:SetBackdrop(backdrop_title)
    titleFrame:Show()

    local closebtn =       _G[frameName .. "_CloseButton"] or 
        CreateFrame("Button", frameName .. "_CloseButton",
        titleFrame, "SackvilleVBCloseButton")
    closebtn:SetPoint("RIGHT", -7, 0)
    closebtn:Show()

    local titleFontString = _G[frameName .. "_TitleText"] or 
        titleFrame:CreateFontString(frameName .. "_TitleText")
    titleFontString:SetFontObject("GameFontNormalLarge")
    titleFontString:SetText(bagName)
    titleFontString:ClearAllPoints()
    titleFontString:SetPoint("RIGHT", -40, 0)
    titleFrame.text = titleFontString
    -- See http://www.wowwiki.com/API_FontInstance_SetFontObject

    titleFrame:SetWidth(titleFontString:GetStringWidth() + 50)
    f:SetWidth(titleFrame:GetWidth())
    f:SetMinResize(titleFrame:GetWidth(), BAG_MIN_HEIGHT)

    f.titleFrame = titleFrame
    
    return f
end

function Sackville.VirtualBag:GetFrame()
    return self.frame
end

function Sackville.VirtualBag:CreateTitleBar() -- UNUSED (for now)
    local basename = "SackvilleVB" .. self.name
    local titleFrame = CreateFrame("Frame", basename .. "titleFrame", self:GetFrame())
    titleFrame:SetPoint("TOPLEFT")
    titleFrame:SetHeight(20)
    titleFrame:SetWidth(self:GetFrame():GetWidth())
    
    local titleFontString = titleFrame:CreateFontString()
    titleFontString:SetText("Sackville")
    titleFontString:SetPoint("CENTER")
end

function Sackville.VirtualBag:AddItem(item, sectionName)
    -- Sackville:Print("AddItem: B:" .. self.name .. "/S:" .. sectionName .. " <- " .. Sackville:DumpItem(item))
    local section = self.sections[sectionName]
    section.items[item.bagslot] = item
    item.removed = false

    item.button.wrapper:SetParent(section.frame)
    item.section = section
    
    if section.frame:IsVisible() then
        PositionItemButton(item)
        UpdateSectionDimensions(section)
    end
    
end

function Sackville.VirtualBag:RemoveItem(item)
    -- Sackville:Print(self.name .. ":RemoveItem: " .. Sackville:DumpItem(item))
    -- Sackville:Print("RemoveItem: B:" .. self.name .. "/S:" .. item.section.name .. " -> " .. Sackville:DumpItem(item))
    item.button.wrapper:Hide()
    item.removed = true
    item.section.items[item.bagslot] = nil
    item.section = nil
end

function Sackville.VirtualBag:LayoutItems()
    -- Sackville:Print("LayoutItems for vbag: " .. self.name)
    -- A section should first try to fit in the remaining width, and then
    -- move below the existing section if that is not possible.
    self.frame:SetWidth(self.min_width)
    self.next_x = self.padding_left
    self.next_y = -self.padding_top
    local remainingWidth = self.max_width - self.padding_left - self.padding_right
    local prevSection = nil

    sectionSortCritera = {
        [1] = {
            key = "displayOrder",
            comparison = "<"
        }
    }
        
    -- for sectionid, section in pairs(self.sections) do
    for sectionName, section in spairs(self.sections, sectionSortCritera) do
        -- Sackville:Print("  Section: " .. section.name)
        if TableHasValues(section.items) then
            section.frame:Show()
            section.row = 0
            section.col = 0
            
            for bagslot, item in kpairs(section.items, section.sortCriteria) do
                local item = section.items[bagslot]

                if GetContainerItemID(item.bagid, item.slot) then -- slot is not empty
                    PositionItemButton(item)
                else    
                    item.button.wrapper:Hide()
                end
                
                -- if item.removed then
                    -- item.button.wrapper:Hide()
                -- else
                    -- PositionItemButton(item)
                -- end
            end
            -- Now all the buttons in the section should be layed out

            if section.frame:GetWidth() <= remainingWidth then
                section.frame:SetPoint("TOPLEFT", self.next_x, self.next_y)
                self.next_x = self.next_x + section.frame:GetWidth() + SECTION_PADDING
                remainingWidth = remainingWidth - section.frame:GetWidth() - SECTION_PADDING
                -- Sackville:Print(string.format("    next_x:%d, remainingWidth: %d",
                    -- self.next_x, remainingWidth))
            else
                remainingWidth = self.max_width - self.padding_left - self.padding_right
                self.next_x = self.padding_left
                self.next_y = self.next_y - prevSection.frame:GetHeight() - SECTION_PADDING
                section.frame:SetPoint("TOPLEFT", self.next_x, self.next_y)
                self.next_x = self.next_x + section.frame:GetWidth() + SECTION_PADDING
                remainingWidth = remainingWidth - section.frame:GetWidth() - SECTION_PADDING
                
                -- Sackville:Print(string.format("next_x:%d, next_y:%d, remainingWidth: %d",
                    -- self.next_x, self_next_y, remainingWidth))
            end
            
            if self.next_x > self.frame:GetWidth() then
                self.frame:SetWidth(self.next_x + self.padding_right)
            end
            
            if self.next_y > section.frame:GetHeight() then
                self.frame:SetHeight(math.abs(self.next_y) + BUTTON_SIZE + self.padding_bottom)
            else
                self.frame:SetHeight(math.abs(self.next_y) + section.frame:GetHeight() + self.padding_bottom)
            end

            prevSection = section        
        else
            section.frame:Hide()
        end
    end

    -- self.frame:SetWidth(math.max(self.frame:GetWidth(),
    self.frame.titleFrame:SetWidth(self.frame:GetWidth())
    
end

function Sackville.VirtualBag:UpdateDimensions() -- not used
    f = self:GetFrame()
       
    if (self.row == 0) then
        f:SetWidth(self.col * BUTTON_SIZE)
    else
        f:SetWidth(
            MAX_WIDTH - (
                (MAX_WIDTH - PADDING_LEFT - PADDING_RIGHT) % BUTTON_SIZE
            )
        )
            
    end
   
    f:SetHeight((1 + self.row) * BUTTON_SIZE + PADDING_TOP + PADDING_BOTTOM)
    f.titleFrame:SetWidth(f:GetWidth())
end

function Sackville.VirtualBag:Show()
    self:LayoutItems()
    self:GetFrame():Show()
end

function Sackville.VirtualBag:Hide()
    self:GetFrame():Hide()
end

function Sackville.VirtualBag:Toggle()
    Sackville:Print("Toggle")
    if (self:GetFrame():IsVisible()) then
        self:Hide()
    else
        self:Show()
    end
end


-- TODO
-- Add vbag footer with currency and free slots
-- Add bank bags
-- support bag minimization
-- support section minimization
-- Add control bar for: search, tidy, clear new, options, show real bags